from syntax_visitor import SyntaxVisitor
from ..ndtypes import ScalarT, TupleT

class Yes(Exception):
  pass 

class ContainsSlices(SyntaxVisitor):
  def visit_Index(self, expr):
    if isinstance(expr.index.type, ScalarT):
      idx_types = [expr.index.type]
    elif isinstance(expr.index.type, TupleT):
      idx_types = expr.index.type.elt_types
    else: 
      raise Yes()
    if not all(isinstance(t, ScalarT) for t in idx_types):
      raise Yes()
  
def contains_slices(fn):
  try:
    ContainsSlices().visit_fn(fn)
    return False 
  except Yes:
    return True


from ..c_backend import prepare_args 
from ..config import stride_specialization 
from ..transforms.pipeline import flatten
from ..transforms.stride_specialization import specialize


from compiler import compile_entry 

def run(fn, args):
  args = prepare_args(args, fn.input_types)
  fn = flatten(fn)
  if stride_specialization:
    fn = specialize(fn, python_values = args)
  compiled_fn = compile_entry(fn)
  assert len(args) == len(fn.input_types)
  result = compiled_fn.c_fn(*args)
  return result


import parakeet
import numpy as np 


def dot(x,y):
  return sum(x*y)

def matmult_high_level(X,Y):
  return np.array([[np.dot(x,y) for y in Y.T] for x in X])

def matmult_loops(X,Y,Z):
  m, d = X.shape
  n = Y.shape[1]
  for i in xrange(m):
    for j in xrange(n):
      total = X[i,0] * Y[0,j] 
      for k in xrange(1,d):
        total += X[i,k] * Y[k,j]
      Z[i,j] = total 
  return Z
     

n, d = 250, 1000
m = 250 
X = np.random.randn(n,d)
Y = np.random.randn(d,m)
Z = np.zeros((m,n))
from timer import compare_perf

compare_perf(matmult_high_level, [X,Y],cpython=True)

compare_perf(matmult_loops, [X, Y, Z], cpython=False)



from .. import names 
from ..builder import build_fn 
from ..ndtypes import Int64, repeat_tuple, NoneType, ScalarT 
from ..syntax import (ParFor, IndexReduce, IndexScan, IndexFilter, Index, Map, OuterMap, 
                      Var, Return, UntypedFn, Expr)
from ..syntax.helpers import unwrap_constant, get_types, none, zero_i64 
from ..syntax.adverb_helpers import max_rank_arg
from inline import Inliner 
from transform import Transform


class IndexifyAdverbs(Transform):
  """
  Take all the adverbs whose parameterizing functions assume they 
  get fed slices of input data and turn them into version which take explicit
  input indices
  """
  
  def fresh_input_name(self, expr):
    if expr is Var:
      return names.refresh(expr.name)
    else:
      return names.fresh("input")
  
  def fresh_fn_name(self, prefix, fn):
    return names.fresh(prefix + names.original(self.get_fn(fn).name))
  
  _indexed_fn_cache = {}
  def indexify_fn(self, fn, axis,
                   array_args, 
                   cartesian_product = False,
                   output = None, 
                   index_offsets = None):
    """
    Take a function whose last k values are slices through input data 
    and transform it into a function which explicitly extracts its arguments
    """  
    array_args = tuple(array_args)
    
    array_arg_types = tuple(get_types(array_args))

    n_arrays = len(array_arg_types)
    closure_args = self.closure_elts(fn)
    closure_arg_types = tuple(get_types(closure_args))
    n_closure_args = len(closure_args)
    fn = self.get_fn(fn)
    
    axes = self.get_axes(array_args, axis)
    
    key = (  fn.cache_key, 
             axes, 
             closure_arg_types, 
             array_arg_types, 
             output is None,  
             cartesian_product, 
             index_offsets, 
           )
    
    def mk_closure():
      new_fn = self._indexed_fn_cache[key] 
      if output is None:
        return self.closure(new_fn, closure_args + array_args)
      else:
        return self.closure(new_fn, (output, ) + closure_args + array_args)
    
    if key in self._indexed_fn_cache:
      return mk_closure()
      
    
    n_indices = n_arrays if cartesian_product else 1
    index_input_type = Int64 if n_indices == 1 else repeat_tuple(Int64, n_arrays) 
    
    if output is None:
      inner_input_types = closure_arg_types + array_arg_types + (index_input_type,)
      new_return_type = fn.return_type 
    else:
      inner_input_types = (output.type,) + closure_arg_types +  array_arg_types + (index_input_type,)
      new_return_type = NoneType 
    
    
    input_names = []
    if output is not None:
      if output is Var:
        local_output_name = names.refresh(output.name)
      else:
        local_output_name = names.fresh("local_output")
      input_names.append(local_output_name) 
    
    for old_input_name in fn.arg_names:
      input_names.append(names.refresh(old_input_name)) 
    
    input_names.append(names.fresh("idx"))
    new_fn_name = self.fresh_fn_name("idx_", fn)
    
    new_fn, builder, input_vars = build_fn(inner_input_types, 
                                           new_return_type,
                                           name = new_fn_name,  
                                           input_names = input_names)

    index_input_var = input_vars[-1]
    if output is None:
      output_var = None
      closure_arg_vars = input_vars[:n_closure_args]
      array_arg_vars = input_vars[n_closure_args:-1]
    else:
      output_var = input_vars[0]
      closure_arg_vars = input_vars[1:n_closure_args+1]
      array_arg_vars = input_vars[n_closure_args+1:-1]
    
    if cartesian_product:
      index_elts = self.tuple_elts(index_input_var) if n_indices > 1 else [index_input_var]
    else:
      assert isinstance(index_input_var.type, ScalarT), \
        "Unexpected index type %s" % (index_input_var.type)
      index_elts = [index_input_var] * n_arrays 
    
    slice_values = []

    for i, curr_array in enumerate(array_arg_vars):
      axis = axes[i]
      idx_expr = index_elts[i]
      if index_offsets is not None:
        assert len(index_offsets) == len(array_arg_vars)
        idx_expr = builder.add(idx_expr, builder.int(index_offsets[i]) )
      curr_slice = builder.slice_along_axis(curr_array, axis, idx_expr)
      
      slice_values.append(curr_slice) 
    
    elt_result = builder.call(fn, tuple(closure_arg_vars) + tuple(slice_values))
    if output is None: 
      builder.return_(elt_result)
    else:
      builder.setidx(output_var, index_input_var, elt_result)
      builder.return_(none)
    
    #inliner = Inliner()
    #new_fn = inliner.apply(new_fn)
    self._indexed_fn_cache[key] = new_fn
    return mk_closure()
          
    
  

  def sizes_along_axis(self, xs, axis):
    axis_sizes = [self.size_along_axis(x, axis)
                  for x in xs
                  if self.rank(x) > axis]

    assert len(axis_sizes) > 0
    # all arrays should agree in their dimensions along the
    # axis we're iterating over
    self.check_equal_sizes(axis_sizes)
    return axis_sizes



  def create_map_output_array(self, fn, array_args, axes, 
                            cartesian_product = False, 
                            name = "output"):
    """
    Given a function and its argument, use shape inference to figure out the
    result shape of the array and preallocate it.  If the result should be a
    scalar, just return a scalar variable.
    """
    assert self.is_fn(fn), \
      "Expected function, got %s" % (fn,)
    assert isinstance(array_args, (list,tuple)), \
      "Expected list of array args, got %s" % (array_args,)
    axes = self.get_axes(array_args, axes)
    
    
    # take the 0'th slice just to have a value in hand 
    inner_args = [self.slice_along_axis(array, axis, zero_i64)
                  for array, axis in zip(array_args, axes)]

    extra_dims = []
    if cartesian_product:
      for array, axis in zip(array_args, axes):
        if self.rank(array) > axis:
          dim = self.shape(array, axis)
        else:
          dim = 1 
        extra_dims.append(dim)
    else:
      extra_dims.append(self.niters(array_args, axes))
    outer_shape_tuple = self.tuple(extra_dims)
    return self.create_output_array(fn, inner_args, outer_shape_tuple, name)

  def get_axes(self, args, axis):
    if isinstance(axis, Expr):
      axis = unwrap_constant(axis)
      
    if axis is None: 
      args = [self.ravel(arg) for arg in args]
      axis = 0
      
    if isinstance(axis, list):
      axes = tuple(axis)
    elif isinstance(axis, tuple):
      axes = axis
    else:
      assert isinstance(axis, (int,long)), "Invalid axis %s" % axis 
      axes = (axis,) * len(args)
       
    assert len(axes) == len(args), "Wrong number of axes (%d) for %d args" % (len(axes), len(args))
    return axes 
  
  def niters(self, args, axes):
    axes = self.get_axes(args, axes)
    assert len(args) == len(axes)
    best_rank = 0 
    best_arg = None
    best_axis = None 
    for curr_arg, curr_axis in zip(args,axes):
      r = self.rank(curr_arg) 
      if r > best_rank:
        best_arg = curr_arg 
        best_axis = curr_axis 
    return self.shape(best_arg, best_axis)
  
  def transform_Map(self, expr, output = None):
    # recursively descend down the function bodies to pull together nested ParFors
    args = self.transform_expr_list(expr.args)
    axes = self.get_axes(args, expr.axis)
    old_fn = expr.fn

    if output is None:
      output = self.create_map_output_array(old_fn, args, axes)
    
    niters = self.niters(args, axes)
    index_fn = self.indexify_fn(expr.fn, axes, args, 
                                cartesian_product=False, 
                                output = output)
    
    
    self.parfor(index_fn, niters)
    return output 
  
  def transform_OuterMap(self, expr):
    args = self.transform_expr_list(expr.args)
    axes = self.get_axes(args, expr.axis)
    
    fn = expr.fn 
    # recursively descend down the function bodies to pull together nested ParFors 
    counts = [self.size_along_axis(arg, axis) for (arg,axis) in zip(args,axes)]
    outer_shape = self.tuple(counts)
    zero = self.int(0)
    first_values = [self.slice_along_axis(arg, axis, zero) for (arg,axis) in zip(args, axes)]
    # self.create_output_array(fn, inner_args, outer_shape, name)
    output =  self.create_output_array(fn, first_values, outer_shape)
    loop_body = self.indexify_fn(fn, axes, args, 
                                 cartesian_product = True, 
                                 output = output)
    self.parfor(loop_body, outer_shape)
    return output 
  
  def transform_IndexMap(self, expr, output = None):
    shape = expr.shape 
    fn = expr.fn 
    
    dims = self.tuple_elts(shape)
    
    if len(dims) == 1:
      shape = dims[0]
    
    if output is None:
      output = self.create_output_array(fn, [shape], shape)
    
    
    old_closure_args = self.closure_elts(fn)
    old_closure_arg_types = get_types(old_closure_args)
    fn = self.get_fn(fn)
    
    closure_arg_names = [self.fresh_input_name(clos_arg) for clos_arg in old_closure_args] 
    new_closure_vars = [Var(name, type=t) 
                        for name, t in 
                        zip(closure_arg_names, old_closure_arg_types)]
    
    idx_name = names.refresh(fn.arg_names[-1])
    output_name = names.refresh("output")  
    
    new_input_names = [output_name] + closure_arg_names + [idx_name]            
    new_input_types =  [output.type]  + old_closure_arg_types + [fn.input_types[-1]]
    new_fn, builder, input_vars = build_fn(new_input_types, NoneType,
                                           name =  names.fresh("idx_" + names.original(fn.name)),  
                                           input_names = new_input_names)
    output_var = input_vars[0]
    idx_var = input_vars[-1]
    builder.setidx(output_var, idx_var, builder.call(fn, new_closure_vars + [idx_var]))
    builder.return_(none)
    new_closure = self.closure(new_fn, (output,) + tuple(old_closure_args)  )
    self.parfor(new_closure, shape)
    return output
  
  
  def transform_Reduce(self, expr):
    fn = expr.fn 
    combine = expr.combine 
    init = expr.init 
    args = expr.args

    axis = expr.axis
    if axis is  None or self.is_none(axis):
      assert len(args) == 1
      args = [self.ravel(args[0])]
      axis = 0
    else:
      axis = unwrap_constant(axis)

    max_arg = max_rank_arg(args)
    nelts = self.shape(max_arg, axis)
    if self.is_none(init):
      assert len(args) == 1, "If 'init' not specified then can't have more than 1 arg"
      init = self.index_along_axis(args[0], axis, self.int(0))
      index_offsets = (1,)
      assert init.type == fn.return_type
      nelts = self.sub(nelts, self.int(1), "nelts") 
    else:
      index_offsets = None
      
    index_fn = self.indexify_fn(fn, 
                                axis, 
                                args, 
                                cartesian_product=False, 
                                index_offsets = index_offsets)


    
    
    return IndexReduce(fn = index_fn, 
                       init = init, 
                       combine = combine, 
                       shape = nelts, 
                       type = expr.type)
   
  def transform_Scan(self, expr, output = None):
    combine = expr.combine 
    init = expr.init 
    args = expr.args

    axis = expr.axis
    if axis is  None or self.is_none(axis):
      assert len(args) == 1
      args = [self.ravel(args[0])]
      axis = 0
    else:
      axis = unwrap_constant(axis)
    
    niters = self.niters(args, axis)
    
    if self.is_none(init):
      assert len(args) == 1
      init = self.index_along_axis(args[0], axis, self.int(0))# self.call(index_fn, [self.int(0)])  
      index_offsets = (1,)
      niters = self.sub(niters, self.int(1), "niters")
    else:
      index_offsets = None
      
      
    index_fn = self.indexify_fn(expr.fn, 
                                axis, 
                                args, 
                                cartesian_product=False, 
                                index_offsets = index_offsets)
    

    return IndexScan(fn = index_fn, 
                     init = init, 
                     combine = combine, 
                     shape = niters,
                     type = expr.type)
  
  def transform_Filter(self, expr):
    assert False, "Filter not implemented"
    
  def transform_Assign(self, stmt):
    """
    If you encounter an adverb being written to an output location, 
    then why not just use that as the output directly? 
    """
    if stmt.lhs.__class__ is Index:
      rhs_class = stmt.rhs.__class__ 
      if rhs_class is Map:
        self.transform_Map(stmt.rhs, output = stmt.lhs)
        return None 
      elif rhs_class is OuterMap:
        self.transform_OuterMap(stmt.rhs, output = stmt.lhs)
       
    return Transform.transform_Assign(self, stmt)

  



##########################
#  Performance Options   #
##########################
fast_math = True
use_openmp = False 

##########################
# Insert Debugging Code  #
##########################
debug = False
check_pyobj_types = False 

#########################
#  Verbose Printing     #
#########################
print_input_ir = False
print_command_elapsed_time = False
print_commands = False
print_function_source = False
print_module_source = False

# Generate a .c file or a .cpp? 
pure_c = True
delete_temp_files = True 




from .. import prims 
from ..frontend import jit, macro
from ..ndtypes import ScalarT, ArrayT, make_array_type
from ..syntax import Select, DelayUntilTyped, PrimCall, Call,  OuterMap


from numpy_reductions import vdot
 
@macro
def _select(cond, trueval, falseval):
  return Select(cond, trueval, falseval)

#@macro  
def where(cond, trueval, falseval):
  #return Select(cond, trueval, falseval )
  return map(_select, cond, trueval, falseval)

@macro
def dot(X,Y):
  def typed_dot(X, Y):
    if isinstance(X.type, ScalarT):
      return PrimCall(prims.multiply, [X, Y], type = Y.type.combine(X.type))
    elif isinstance(Y.type, ScalarT):
      return PrimCall(prims.multiply, [X, Y], type = X.type.combine(Y.type))
    else:
      assert isinstance(X.type, ArrayT), \
        "Expected %s to be array but got %s" % (X, X.type)
      assert isinstance(X.type, ArrayT), \
        "Expected %s to be array but got %s" % (Y, Y.type)
      
      from ..frontend import ast_conversion 
      vdot_untyped = ast_conversion.translate_function_value(vdot)
      from ..type_inference import specialize 
      vdot_typed = specialize(vdot_untyped, 
                              [make_array_type(X.type.elt_type, 1), 
                               make_array_type(Y.type.elt_type, 1)])
      result_type = X.type.elt_type.combine(Y.type.elt_type)
      assert vdot_typed.return_type == result_type, \
        "Expected return type %s but got %s from vdot" % (result_type, vdot_typed.return_type)
      
      if X.type.rank == 1 and Y.type.rank == 1:
        return Call(fn = vdot_untyped, args = [X, Y], type = result_type)

      if X.type.rank == 1:
        assert Y.type.rank == 2, "Don't know how to multiply %s and %s" % (X.type, Y.type)

        assert False, "Dot product between vector and matrix not yet implemented"
        
      elif Y.type.rank == 1:
        assert X.type.rank == 1, "Don't know how to multiply %s and %s" % (X.type, Y.type)
        assert False, "Dot product between matrix and vector not yet implemented"
        
      assert X.type.rank == 2 and Y.type.rank == 2, \
        "Don't know how to multiply %s and %s" % (X.type, Y.type)
      
      return OuterMap(fn = vdot_typed, args = (X, Y), axis = (0,1), 
                      type = make_array_type(result_type, 2))
      
  return DelayUntilTyped(  [X, Y], typed_dot)
    


 
import ctypes

from treelike import NestedBlocks

from .. import names, prims 
from ..analysis import use_count
from ..syntax import (Const, Tuple, TypedFn, Var, TupleProj, ArrayView, 
                      PrimCall, Attribute, Expr, Closure)  
from ..ndtypes import (TupleT,  ArrayT, NoneT, 
                       elt_type, ScalarT, 
                       FloatT, Float32, Float64, 
                       BoolT, Bool, 
                       IntT,  Int64, SignedT,
                       PtrT, NoneType, 
                       FnT, ClosureT) 


from c_types import to_ctype, to_dtype
from base_compiler import BaseCompiler

from compile_util import compile_module
from config import (debug, 
                    check_pyobj_types, 
                    print_function_source, 
                    print_module_source, 
                    print_input_ir, 
                    ) 

def compile_flat_source(fn, _compile_cache = {}):
  key = fn.cache_key
  if key in _compile_cache:
    return _compile_cache[key]
  compiler = FlatFnCompiler()
  name, sig, src = compiler.visit_fn(fn)
  _compile_cache[key] = (name,sig,src)
  return (name,sig,src)


def compile_entry(fn, _compile_cache = {}):
  key = fn.cache_key
  if key in _compile_cache:
    return _compile_cache[key]
  compiler = PyModuleCompiler()
  name, sig, src = compiler.visit_fn(fn)
  if print_function_source: print "Generated C source for %s: %s" %(name, src)
  compiled_fn = compile_module(src, 
                               fn_name = name,
                               fn_signature = sig, 
                               extra_objects = set(compiler.extra_objects),
                               extra_function_sources = compiler.extra_function_sources, 
                               forward_declarations =  compiler.forward_declarations, 
                               print_source = print_module_source)
  _compile_cache[key]  = compiled_fn
  return compiled_fn


def entry_function_source(fn):
  return compile_entry(fn).src 

def entry_function_name(fn):
  return compile_entry(fn).fn_name 

def entry_function_signature(fn):
  return compile_entry(fn).fn_signature 

class FlatFnCompiler(BaseCompiler):
  
  def __init__(self):
    BaseCompiler.__init__(self)
    
    self.forward_declarations = set([])
    # depends on these .o files
    self.extra_objects = set([]) 
    
    # to avoid adding the same function's source twice 
    # we also track the signatures in a set 
    self.extra_function_signatures = set([])
    self.extra_function_sources = []
    
  def visit_Alloc(self, expr):
    elt_t =  expr.elt_type
    nelts = self.fresh_var("npy_intp", "nelts", self.visit_expr(expr.count))

    return "(PyArrayObject*) PyArray_SimpleNew(1, &%s, %s)" % (nelts, to_dtype(elt_t))
  
  
  
  def visit_Const(self, expr):
    if isinstance(expr.type, BoolT):
      return "1" if expr.value else "0"
    elif isinstance(expr.type, NoneT):
      return "0"
    return "%s" % expr.value 
  
  def visit_Var(self, expr):
    return self.name(expr.name)
  
  def visit_Cast(self, expr):
    x = self.visit_expr(expr.value)
    ct = to_ctype(expr.type)
    if isinstance(expr, (Const, Var)):
      return "(%s) %s" % (ct, x)
    else:
      return "((%s) (%s))" % (ct, x)
  
  
  def not_(self, x):
    if x == "1":
      return "0"
    elif x == "0":
      return "1"
    return "!%s" % x
  
  def and_(self, x, y):
    if x == "0" or y == "0":
      return "0"
    elif x == "1" and y == "1":
      return "1"
    elif x == "1":
      return y 
    elif y == "1":
      return x
    return "%s && %s" % (x,y) 
  
  def or_(self, x, y):
    if x == "1" or y == "1":
      return "1"
    elif x == "0":
      return y
    elif y == "0":
      return x 
    return "%s || %s" % (x,y) 
  
  def visit_PrimCall(self, expr):
    t = expr.type
    args = self.visit_expr_list(expr.args)
    
    # parenthesize any compound expressions 
    for i, arg_expr in enumerate(expr.args):
      if not isinstance(arg_expr, (Var, Const)):
        args[i] = "(" + args[i] + ")"
        
    p = expr.prim 
    if p == prims.add:
      return "%s + %s" % (args[0], args[1])
    if p == prims.subtract:
      return "%s - %s" % (args[0], args[1])
    elif p == prims.multiply:
      return "%s * %s" % (args[0], args[1])
    elif p == prims.divide:
      return "%s / %s" % (args[0], args[1])
    elif p == prims.negative:
      if t == Bool:
        return "1 - %s" % args[0]
      else:
        return "-%s" % args[0]
    elif p == prims.abs:
      x  = args[0]
      return " %s >= 0 ? %s  : -%s" % (x,x,x)
    
    elif p == prims.bitwise_and:
      return "%s & %s" % (args[0], args[1])
    elif p == prims.bitwise_or:
      return "%s | %s" % (args[0], args[1])
    elif p == prims.bitwise_not:
      return "~%s" % args[0]
    
    elif p == prims.logical_and:
      return self.and_(args[0], args[1])
      
    elif p == prims.logical_or:
      return self.or_(args[0], args[1])
    
    elif p == prims.logical_not:
      return self.not_(args[0])
      
    elif p == prims.equal:
      if isinstance(t, (BoolT, IntT)) and args[0] == args[1]:
        return "1"
      return "%s == %s" % (args[0], args[1])
    elif p == prims.not_equal:
      if isinstance(t, (BoolT, IntT)) and args[0] == args[1]:
        return "0"
      return "%s != %s" % (args[0], args[1])
    elif p == prims.greater:
      if isinstance(t, (BoolT, IntT)) and args[0] == args[1]:
        return "0"
      return "%s > %s" % (args[0], args[1])
    elif p == prims.greater_equal:
      if isinstance(t, (BoolT, IntT)) and args[0] == args[1]:
        return "1"
      return "%s >= %s" % (args[0], args[1])
    elif p == prims.less:
      if isinstance(t, (BoolT, IntT)) and args[0] == args[1]:
        return "0"
      return "%s < %s" % (args[0], args[1])
    elif p == prims.less_equal:
      if isinstance(t, (BoolT, IntT)) and args[0] == args[1]:
        return "1"
      return "%s <= %s" % (args[0], args[1])
    elif p == prims.remainder:
      x,y = args
      if t == Float32: return "fmod(%s, %s)" % (x,y)
      elif t == Float64: return "fmod(%s, %s)" % (x,y)
      assert isinstance(t, (BoolT, IntT)), "Modulo not implemented for %s" % t
      rem = self.fresh_var(t, "rem", "%s %% %s" % (x,y))
      y_is_negative = self.fresh_var(t, "y_is_negative", "%s < 0" % y)
      rem_is_negative = self.fresh_var(t, "rem_is_negative", "%s < 0" % rem)
      y_nonzero = self.fresh_var(t, "y_nonzero", "%s != 0" % y)
      rem_nonzero = self.fresh_var(t, "rem_nonzero", "%s != 0" % rem)
      neither_zero = self.fresh_var(t, "neither_zero", "%s && %s" % (y_nonzero, rem_nonzero))
      diff_signs = self.fresh_var(t, "diff_signs", "%s ^ %s" % (y_is_negative, rem_is_negative))
      should_flip = self.fresh_var(t, "should_flip", "%s && %s" % (neither_zero, diff_signs))
      flipped_rem = self.fresh_var(t, "flipped_rem", "%s + %s" % (y, rem))
      return "%s ? %s : %s" % (should_flip, flipped_rem, rem)
    elif p == prims.fmod:
      if t == Float32: return "fmodf(%s, %s)" % (args[0], args[1])
      elif t == Float64: return "fmod(%s, %s)" % (args[0], args[1])
      return "%s %% %s" % (args[0], args[1])
    elif p == prims.maximum:
      x,y = args
      return "(%s > %s) ? %s : %s" % (x,y,x,y)
    elif p == prims.minimum:
      x,y = args
      return "(%s < %s) ? %s : %s" % (x,y,x,y)
    
    elif p == prims.power:
      if t == Float32: 
        return "powf(%s, %s)" % (args[0], args[1])
      else:
        return "pow(%s, %s)" % (args[0], args[1])
    
    elif isinstance(t, FloatT):
      # many float prims implemented using the same name in math.h
      name = p.name
      if name.startswith("arc"):
        # arccos -> acos
        name = "a" + name[3:]
      if t == Float32: name = name + "f" 
      if len(args) == 1:
        return "%s(%s)" % (name, args[0])
      else:
        assert len(args) == 2, "Unexpected prim %s with %d args (%s)" % (p, len(args), args)
        return "%s(%s, %s)" % (name, args[0], args[1])
  
    else:
      assert False, "Prim not yet implemented: %s" % p
  
  def visit_Index(self, expr):
    arr = self.visit_expr(expr.value)
    idx = self.visit_expr(expr.index)
    elt_t = expr.value.type.elt_type
    ptr_t = "%s*" % to_ctype(elt_t)
    return "( (%s) (PyArray_DATA(%s)))[%s]" % (ptr_t, arr, idx)
  
  def visit_Call(self, expr):
    fn_name = self.get_fn(expr.fn)
    closure_args = self.get_closure_args(expr.fn)
    args = self.visit_expr_list(expr.args)
    return "%s(%s)" % (fn_name, ", ".join(tuple(closure_args) + tuple(args)))
  
  def visit_Select(self, expr):
    cond = self.visit_expr(expr.cond)
    true = self.visit_expr(expr.true_value)
    false = self.visit_expr(expr.false_value)
    return "%s ? %s : %s" % (cond, true, false) 
  
  def is_pure(self, expr):
    return expr.__class__ in (Var, Const, PrimCall, Attribute, TupleProj, Tuple, ArrayView)
  
  def visit_Assign(self, stmt):
    lhs = self.visit_expr(stmt.lhs)
    rhs = self.visit_expr(stmt.rhs)
    
    if stmt.lhs.__class__ is Var:
      return "%s %s = %s;" % (to_ctype(stmt.lhs.type), lhs, rhs)
    else:
      return "%s = %s;" % (lhs, rhs)
  
  def declare(self, parakeet_name, parakeet_type, init_value = None):
    c_name = self.name(parakeet_name)
    t = to_ctype(parakeet_type)
    if init_value is None:
      self.append("%s %s;" % (t, c_name))
    else: 
      self.append("%s %s = %s;" % (t, c_name, init_value))
  
  def declare_merge_vars(self, merge):
    """ 
    Declare but don't initialize
    """
    for (name, (left, _)) in merge.iteritems():
      self.declare(name, left.type)
      
  def visit_merge_left(self, merge, fresh_vars = True):
    
    if len(merge) == 0:
      return ""
    
    self.push()
    self.comment("Merge Phi Nodes (left side) " + str(merge))
    for (name, (left, _)) in merge.iteritems():
      c_left = self.visit_expr(left)
      if fresh_vars:
        self.declare(name, left.type, c_left)
      else:
        c_name = self.name(name)
        self.append("%s = %s;" % (c_name, c_left))
        
    return self.pop()
  
  def visit_merge_right(self, merge):
    
    if len(merge) == 0:
      return ""
    self.push()
    self.comment("Merge Phi Nodes (right side) " + str(merge))
    
    for (name, (_, right)) in merge.iteritems():
      c_right = self.visit_expr(right)
     
      self.append("%s = %s;"  % (self.name(name), c_right))
    return self.pop()
  
  def visit_If(self, stmt):
    self.declare_merge_vars(stmt.merge)
    cond = self.visit_expr(stmt.cond)
    true = self.visit_block(stmt.true) + self.visit_merge_left(stmt.merge, fresh_vars = False)
    false = self.visit_block(stmt.false) + self.visit_merge_right(stmt.merge)
    return self.indent("if(%s) {\n%s\n} else {\n%s\n}" % (cond, self.indent(true), self.indent(false))) 
  
  def visit_While(self, stmt):
    decls = self.visit_merge_left(stmt.merge, fresh_vars = True)
    cond = self.visit_expr(stmt.cond)
    body = self.visit_block(stmt.body) + self.visit_merge_right(stmt.merge)
    return decls + "while (%s) {%s}" % (cond, body)
  
  def visit_ForLoop(self, stmt):
    s = self.visit_merge_left(stmt.merge, fresh_vars = True)
    start = self.visit_expr(stmt.start)
    stop = self.visit_expr(stmt.stop)
    step = self.visit_expr(stmt.step)
    var = self.visit_expr(stmt.var)
    t = to_ctype(stmt.var.type)
    body =  self.visit_block(stmt.body)
    body += self.visit_merge_right(stmt.merge)
    body = self.indent("\n" + body) 
    
    
    s += "\n %(t)s %(var)s;"
    s += "\nfor (%(var)s = %(start)s; %(var)s < %(stop)s; %(var)s += %(step)s) {%(body)s}"
    return s % locals()

  def visit_Return(self, stmt):
    assert not self.return_by_ref, "Returning multiple values not yet implemented: %s" % stmt

    if self.return_void:
      return "return;"
    else:
      v = self.visit_expr(stmt.value)
      return "return %s;" % v
      
  def visit_block(self, stmts, push = True):
    if push: self.push()
    for stmt in stmts:
      s = self.visit_stmt(stmt)
      self.append(s)
    self.append("\n")
    return self.indent("\n" + self.pop())
  
  def tuple_to_var_list(self, expr):
    assert isinstance(expr, Expr)
    if isinstance(expr, Tuple):
      elts = expr.elts 
    else:
      assert isinstance(expr.type, ScalarT), "Unexpected expr %s : %s" % (expr, expr.type)
      elts = [expr]
    return self.visit_expr_list(elts)
      
  
  def get_fn(self, expr):
    if expr.__class__ is  TypedFn:
      fn = expr 
    elif expr.__class__ is Closure:
      fn = expr.fn 
    else:
      assert isinstance(expr.type, (FnT, ClosureT)), \
        "Expected function or closure, got %s : %s" % (expr, expr.type)
      fn = expr.type.fn
    #compiled_fn = compile_flat(result)
    #self.extra_objects.add(compiled_fn.object_filename)
    #self.forward_declarations.add(compiled_fn.fn_signature)
    c_name, sig, src = compile_flat_source(fn)
    if sig not in self.extra_function_signatures:
      self.extra_function_signatures.add(sig)
      self.extra_function_sources.append(src)
    return c_name

  def get_closure_args(self, fn):
    if isinstance(fn.type, FnT):
      return []
    else:
      assert isinstance(fn, Closure), "Expected closure, got %s : %s" % (fn, fn.type)
      return self.visit_expr_list(fn.args)
      
  def build_loops(self, loop_vars, bounds, body):
    if len(loop_vars) == 0:
      return body
    var = loop_vars[0]
    bound = bounds[0]
    nested = self.build_loops(loop_vars[1:], bounds[1:], body)
    return """
    for (%s = 0; %s < %s; ++%s) {
      %s
    }""" % (var, var, bound, var, nested )
    

      
  def visit_TypedFn(self, expr):
    return self.get_fn(expr)

  def visit_UntypedFn(self, expr):
    assert False, "Unexpected UntypedFn %s in C backend, should have been specialized" % expr.name
  
  
  def return_types(self, fn):
    if isinstance(fn.return_type, TupleT):
      return fn.return_type.elt_types
    elif isinstance(fn.return_type, NoneT):
      return []
    else:
      assert isinstance(fn.return_type, (PtrT, ScalarT))
      return [fn.return_type]
    
  
  def visit_fn(self, fn):
    
    c_fn_name = self.fresh_name(fn.name)
    arg_types = [to_ctype(t) for t in fn.input_types]
    arg_names = [self.name(old_arg) for old_arg in fn.arg_names]
    return_types = self.return_types(fn)
    n_return = len(return_types)
    
    if n_return == 1:
      return_type = to_ctype(return_types[0])
      self.return_void = (return_type == NoneType)
      self.return_by_ref = False
    elif n_return == 0:
      return_type = "void"
      self.return_void = True
      self.return_by_ref = False
    else:
      return_type = "void"
      self.return_void = True
      self.return_by_ref = True
      self.return_var_types = [to_ctype(t) for t in return_types]
      self.return_var_names = [self.fresh_name("return_value%d" % i) for i in xrange(n_return)]
      arg_types = arg_types + ["%s*" % t for t in self.return_var_types] 
      arg_names = arg_names + self.return_var_names
      
    args_str = ", ".join("%s %s" % (t, name) for (t,name) in zip(arg_types,arg_names))
    
    body_str = self.visit_block(fn.body) 

    sig = "%s %s(%s)" % (return_type, c_fn_name, args_str)
    src = "%s { %s }" % (sig, body_str) 
    return c_fn_name, sig, src
    

class PyModuleCompiler(FlatFnCompiler):
   
  def unbox_scalar(self, x, t, target = "scalar_value"):
    assert isinstance(t, ScalarT), "Expected scalar type, got %s" % t
    
    result = self.fresh_var(t, target)
    if isinstance(t, IntT):
      check = "PyInt_Check"
      if isinstance(t, SignedT):
        get = "PyInt_AsLong"
      else:
        get = "PyInt_AsUnsignedLongMask"
    elif isinstance(t, FloatT):
      check = "PyFloat_Check"
      get = "PyFloat_AsDouble"
    else:
      assert isinstance(t, BoolT), "Unexpected type %s" % t 
      check = "PyBool_Check"
      get = "PyObject_IsTrue"
    
    self.append("""
      if (%(check)s(%(x)s)) { %(result)s = %(get)s(%(x)s); }
      else { PyArray_ScalarAsCtype(%(x)s, &%(result)s); }
    """ % locals())
    return result 
      
  def box_scalar(self, x, t):  
    if isinstance(t, BoolT):
      return "PyBool_FromLong(%s)" % x
    elif isinstance(t, NoneT):
      self.append("Py_INCREF(Py_None);")
      return "Py_None"
    if x.replace("_", "").isalpha():
      scalar = x
    else:
      scalar = self.fresh_name("scalar");
      self.append("%s %s = %s;" % (to_ctype(t), scalar, x))
    return "PyArray_Scalar(&%s, PyArray_DescrFromType(%s), NULL)" % (scalar, to_dtype(t) )
    
  def as_pyobj(self, expr):
    """
    Compile the expression and if necessary box it up as a PyObject
    """
    x = self.visit_expr(expr)
    if isinstance(expr.type, (NoneT, ScalarT)):
      return self.box_scalar(x, expr.type)
    elif isinstance(expr.type, ArrayT):
      return "(PyObject*) " + x 
    else:
      return x
  
  def as_pyobj_list(self, exprs):
    return [self.as_pyobj(expr) for expr in exprs]
  
  def c_str(self, obj):
    return "PyString_AsString(PyObject_Str(%s))" % obj
  
  def c_type_str(self, obj):
    return self.c_str("PyObject_Type((PyObject*) %s)" % obj)
   
  def print_pyobj(self, obj, text = ""):
    text = '"%s"' % text
    self.append('printf("%%s%%s\\n", %s, %s);' % (text, self.c_str(obj)))
  
  def print_pyobj_type(self, obj, text=""):
    text = '"%s"' % text
    self.append('printf("%%s%%s\\n", %s, %s);' % (text, self.c_type_str(obj)))
    
  def tuple_to_stack_array(self, expr, name = "array_from_tuple", elt_type = None):
    t0 = expr.type.elt_types[0]
    
    assert expr.type.__class__ is TupleT 
    assert all(t == t0 for t in expr.type.elt_types[1:])
    
    if expr.__class__ is Tuple:
      elts = [self.visit_expr(elt_expr) for elt_expr in expr.elts]
    else:
      tup = self.visit_expr(expr)
      self.check_tuple(tup)
      elts = self.tuple_elts(tup, expr.type.elt_types)
    
    array_name = self.fresh_name(name)
    n = len(expr.type.elt_types)
    if elt_type is None:
      elt_type = to_ctype(t0)
    self.append("%s %s[%d];" % (elt_type, array_name, n))
    for i, elt in enumerate(elts):
      self.append("%s[%d] = %s;" % (array_name, i, elt))
    return array_name
    
  def array_to_tuple(self, arr, n, elt_t):
    if n == 0: return "PyTuple_Pack(0);"
    elts = [self.box_scalar("%s[%d]" % (arr,i), elt_t) for i in xrange(n)]
    elt_str = ", ".join(elts)
    return "PyTuple_Pack(%d, %s)" % (n, elt_str)
  
  def tuple_elts(self, tup, ts):
    result = []
    for i,t in enumerate(ts):
      result.append(self.tuple_elt(tup, i, t))
    return result
  
  def mk_tuple(self, elts):
    n = len(elts)
    if n == 0: return "PyTuple_Pack(0);"
    elt_str = ", ".join(self.as_pyobj_list(elts)) 
    return "PyTuple_Pack(%d, %s)" % (n, elt_str)
  
  
  def check_tuple(self, tup):
    if not check_pyobj_types: return 
    self.newline()
    self.comment("Checking tuple type for %s" % tup)
    self.append("""
      if (!PyTuple_Check(%s)) { 
        PyErr_Format(PyExc_AssertionError, 
                    "Expected %s to be tuple, got %%s", 
                    %s); 
        return NULL;
      }""" % (tup, tup, self.c_type_str(tup)))
 
  def check_slice(self, obj):
    if not check_pyobj_types: return 
    self.newline()
    self.comment("Checking slice type for %s" % obj)
    self.append("""
      if (!PySlice_Check(%s)) { 
        PyErr_Format(PyExc_AssertionError, 
                    "Expected %s to be slice, got %%s", 
                    %s); 
        return NULL;
      }""" % (obj, obj, self.c_type_str(obj)))
  
  def check_array(self, arr):
    if not check_pyobj_types: return 
    self.newline()
    self.comment("Checking array type for %s" % arr)
    self.append("""
      if (!PyArray_Check(%s)) { 
        PyErr_Format(PyExc_AssertionError, 
                    "Expected %s to be array, got %%s : %%s", 
                    %s, %s); 
        return NULL;
      }""" % (arr, arr, self.c_str(arr), self.c_type_str(arr)))
  
  
  def check_bool(self, x):
    if not check_pyobj_types: return 
    self.newline()
    self.comment("Checking bool type for %s" % x)
    self.append("""
      if (!PyArray_IsScalar(%s, Bool)) { 
        PyErr_Format(PyExc_AssertionError, 
                     "Expected %s to be bool, got %%s", 
                     %s); 
        return NULL;
      }""" % (x, x, self.c_type_str(x)))
  
  def check_int(self, x):
    if not check_pyobj_types: return 
    self.newline()
    self.comment("Checking int type for %s" % x)
    self.append("""
      if (!PyArray_IsIntegerScalar(%s)) { 
        PyErr_Format(PyExc_AssertionError, 
                     "Expected %s to be int, got %%s", 
                     %s); 
        return NULL;
      }""" % (x, x, self.c_type_str(x)))
  
  def check_type(self, v, t):
    if not check_pyobj_types: return 
    if isinstance(t, (ClosureT, TupleT)):
      self.check_tuple(v)
    elif isinstance(t, BoolT):
      self.check_bool(v)
    elif isinstance(t, IntT):
      self.check_int(v)
    elif isinstance(t, ArrayT):
      self.check_array(v)
      
  def tuple_elt(self, tup, idx, t):
    self.check_tuple(tup)
    proj_str = "PyTuple_GetItem(%s, %d)" % (tup, idx)
    if isinstance(t, ScalarT):
      elt_obj = self.fresh_var("PyObject*", "%s_elt" % tup, proj_str)
      result = self.unbox_scalar(elt_obj, t)
      if debug and t == Int64:
        self.append(""" printf("tupleproj %s[%d] = %%" PRId64 "\\n", %s);""" % (tup, idx, result))
      return result
    else:
      return proj_str 
 
  
  def strides(self, array_expr):
    arr_t = array_expr.type
    assert isinstance(arr_t, ArrayT), \
      "Can only get strides of array, not %s : %s" % (array_expr, arr_t)
    elt_t = arr_t.elt_type
    arr = self.visit_expr(array_expr)
    
    bytes_per_elt = elt_t.dtype.itemsize

    strides_tuple_t = arr_t.strides_t
    stride_t = strides_tuple_t.elt_types[0]
    
    assert all(t == stride_t for t in strides_tuple_t)
    n = len(strides_tuple_t.elt_types)
    strides_bytes = self.fresh_name("strides_bytes")
    self.append("npy_intp* %s = PyArray_STRIDES( (PyArrayObject*) %s);" % (strides_bytes, arr))
    strides_elts = self.fresh_name("strides_elts")
    self.append("npy_intp %s[%d];" % (strides_elts, n))
    for i in xrange(n):
      if debug:
        self.printf("converting strides %s[%d] = %%ld to %%ld" % (strides_bytes, i), 
                    "%s[%d]" % (strides_bytes, i), "%s[%d] / %d" % (strides_bytes, i, bytes_per_elt))   
      self.append("%s[%d] = %s[%d] / %d;" % (strides_elts, i, strides_bytes, i, bytes_per_elt))
    strides_tuple = self.array_to_tuple(strides_elts, n, stride_t)
    return strides_tuple
    
  def decref(self, obj):
    self.append("Py_DECREF(%s);" % obj)
    
  def attribute(self, v, attr, t):
    if attr == "data":
      self.check_array(v)

      result = "(PyArrayObject*) PyArray_Ravel((PyArrayObject*) %s, 0)" % (v,)

      
      return result   

    elif attr == "shape":
      self.check_array(v)
      elt_types = t.elt_types
      n = len(elt_types)
      elt_t = elt_types[0]
      assert all(t == elt_t for t in elt_types)
      shape_name = self.fresh_name("strides")
      shape_array = "PyArray_DIMS( (PyArrayObject*) %s)" % v
      self.append("npy_intp* %s = %s;" % (shape_name, shape_array))
      return self.array_to_tuple(shape_name, n, elt_t)
      
    elif attr == "strides":
      assert False, "Can't directly use NumPy strides without dividing by itemsize"
      
    elif attr == 'offset':
      return "0"
    elif attr in ('size', 'nelts'):
      return "PyArray_Size(%s)" % v
    
    elif attr in ('start', 'stop', 'step'):
      self.check_slice(v)
      obj = "((PySliceObject*)%s)->%s" % (v, attr)
      return self.unbox_scalar(obj, t, attr)
    else:
      assert False, "Unsupported attribute %s" % attr 
   
    
  
  def visit_AllocArray(self, expr):
    shape = self.tuple_to_stack_array(expr.shape)
    t = to_dtype(elt_type(expr.type))
    return "(PyArrayObject*) PyArray_SimpleNew(%d, %s, %s)" % (expr.type.rank, shape, t)
    
  def visit_Tuple(self, expr):
    return self.mk_tuple(expr.elts)
  
  def visit_Closure(self, expr):
    return self.mk_tuple(expr.args)
  
  def visit_TupleProj(self, expr):
    tup = self.visit_expr(expr.tuple)
    result = self.tuple_elt(tup, expr.index, expr.type)
    return result
  
  def visit_ClosureElt(self, expr):
    clos = self.visit_expr(expr.closure)
    return self.tuple_elt(clos, expr.index, expr.type)
  
  
  def visit_ArrayView(self, expr):
    ndims = expr.type.rank 
    vec = self.visit_expr(expr.data)
    offset = self.visit_expr(expr.offset)
    
    if expr.strides.__class__ is Tuple:
      strides_elts = self.visit_expr_list(expr.strides.elts)  
    else:
      strides_var = self.visit_expr(expr.strides)
      strides_array = self.tuple_to_stack_array(strides_var, "strides_array", "npy_intp")
      strides_elts = ["%s[%d]" % (strides_array, i) for i in xrange(ndims)]
    
    if expr.shape.__class__ is Tuple:
      shape_elts = self.visit_expr_list(expr.shape.elts) 
    else:
      shape_var = self.visit_expr(expr.strides)
      shape_array = self.tuple_to_stack_array(shape_var, "shape_array", "npy_intp")
      shape_elts = ["%s[%d]" % (shape_array, i) for i in xrange(ndims)]
    
    # slice out the 1D data array if there's an offset 
    size =  "PySequence_Size( (PyObject*) %(vec)s)" % locals()

    self.append("""
      if (%(offset)s > 0) {
        %(vec)s = (PyArrayObject*) PySequence_GetSlice( (PyObject*)  %(vec)s, %(offset)s, %(size)s);
      }""" % locals())
    
    self.check_array(vec)

    count = self.fresh_var("int64_t", "count", "PySequence_Size( (PyObject*) %s)" % vec)

    
      
    
    # increase the rank of the 1D nd-array to whatever the rank of the result 
    # should be so that the strides and shape arrays are of the appropriate size 
    if ndims > 1:
  
      uprank_elts =  (count,) + ("1",) * (ndims-1) 
      uprank_elts_as_pyobj = ["PyInt_FromLong(%s)"  % elt for elt in uprank_elts]
      uprank_elts_str = ", ".join(uprank_elts_as_pyobj)
      uprank_shape = "PyTuple_Pack(%d, %s)" % (ndims, uprank_elts_str)
      self.append("%s = (PyArrayObject*) PyArray_Reshape( (PyArrayObject*) %s, %s);" % (vec, vec, uprank_shape))
      self.return_if_null(vec)
    numpy_strides = self.fresh_var("npy_intp*", "numpy_strides")
    self.append("%s = PyArray_STRIDES(  (PyArrayObject*) %s);" % (numpy_strides, vec))
    numpy_shape = self.fresh_var("npy_intp*", "numpy_shape")
    self.append("%s = PyArray_DIMS(  (PyArrayObject*) %s);" % (numpy_shape, vec))
    bytes_per_elt = expr.type.elt_type.dtype.itemsize
    
    for i, _ in enumerate(expr.strides.type.elt_types):
      self.append("%s[%d] = %s * %d;" % (numpy_strides, i, strides_elts[i], bytes_per_elt) )
      self.append("%s[%d] = %s;" % (numpy_shape, i, shape_elts[i]))
    
    self.append("""
      // clear both fortran and c layout flags 
      ((PyArrayObject*) %(vec)s)->flags &= ~NPY_F_CONTIGUOUS;
      ((PyArrayObject*) %(vec)s)->flags &= ~NPY_C_CONTIGUOUS;
    """ % locals())
    
    f_layout_strides = ["1"]
    for shape_elt in shape_elts[1:]:
      f_layout_strides.append(f_layout_strides[-1] + " * " + shape_elt)
    
    c_layout_strides = ["1"]
    for shape_elt in list(reversed(shape_elts))[:-1]:
      c_layout_strides = [c_layout_strides[-1] + " * " + shape_elt] + c_layout_strides
    
    
    is_c_layout = "&& ".join("(%s) == (%s)" % (actual, ideal) 
                             for actual, ideal 
                             in zip(strides_elts, c_layout_strides))
    is_f_layout = " && ".join("(%s) == (%s)" % (actual, ideal) 
                             for actual, ideal 
                             in zip(strides_elts, f_layout_strides))
    
       
    # make sure the contiguity flags are set correctly 
    self.append("""
      // it's possible that *neither* of the above flags should be on
      // which is why we enable them separately here 
      if (%(is_f_layout)s) { ((PyArrayObject*)%(vec)s)->flags |= NPY_F_CONTIGUOUS; }
      if (%(is_c_layout)s) { ((PyArrayObject*)%(vec)s)->flags |= NPY_C_CONTIGUOUS; }
    """ % locals())
    return vec
  
    
  def visit_Attribute(self, expr):
    attr = expr.name
    if attr == 'strides':
      return self.strides(expr.value)
    v = self.visit_expr(expr.value) 
    return self.attribute(v, attr, expr.type)
  
  
    
  def visit_Return(self, stmt):
    v = self.as_pyobj(stmt.value)
    if debug: 
      self.print_pyobj_type(v, "Return type: ")
      self.print_pyobj(v, "Return value: ")

    return "return %s;" % v
  
  def visit_block(self, stmts, push = True):
    if push: self.push()
    for stmt in stmts:
      s = self.visit_stmt(stmt)
      self.append(s)
    self.append("\n")
    return self.pop()
      
  def visit_TypedFn(self, expr):
    return self.get_fn(expr)

  def visit_UntypedFn(self, expr):
    assert False, "Unexpected UntypedFn %s in C backend, should have been specialized" % expr.name
  
         
  def visit_fn(self, fn):
    if print_input_ir:
      print "=== Compiling to C (entry function) ==="
      print fn
    c_fn_name = self.fresh_name(fn.name)
    uses = use_count(fn)
    self.push()
    
    
    dummy = self.fresh_name("dummy")
    args = self.fresh_name("args")
    
    if debug: 
      self.newline()
      self.printf("\\nStarting %s : %s..." % (c_fn_name, fn.type))
      
    for i, argname in enumerate(fn.arg_names):
      assert argname in uses, "Couldn't find arg %s in use-counts" % argname
      if uses[argname] <= 1:
        self.comment("Skipping unused argument %s" % argname)
        continue
      self.comment("Unpacking argument %s"  % argname)
      c_name = self.name(argname)
      self.append("PyObject* %s = PyTuple_GetItem(%s, %d);" % (c_name, args, i))
      t = fn.type_env[argname]
      self.check_type(c_name, t)
      if debug:
        self.printf("Printing arg #%d %s" % (i,c_name))
        self.print_pyobj_type(c_name, text = "Type: ")
        self.print_pyobj(c_name, text = "Value: ")
      
      if isinstance(t, ScalarT):
        new_name = self.name(argname, overwrite = True)
        self.unbox_scalar(c_name, t, target = new_name)

        
    c_body = self.visit_block(fn.body, push=False)
    c_body = self.indent("\n" + c_body )#+ "\nPyGILState_Release(gstate);")
    c_args = "PyObject* %s, PyObject* %s" % (dummy, args) #", ".join("PyObject* %s" % self.name(n) for n in fn.arg_names)
    c_sig = "PyObject* %(c_fn_name)s (%(c_args)s)" % locals() 
    fndef = "%s {%s}" % (c_sig, c_body)
    return c_fn_name, c_sig, fndef 

  
    
    
    
    
    


default_backend = 'c' #llvm


######################################
#        PARAKEET OPTIMIZATIONS      #
######################################
  
default_opt_level = 2 


def set_opt_level(n):
  assert 0 <= n <= 3, "Invalid optimization level %d" % n
  g = globals()

  for name in ('opt_inline', 
               'opt_fusion', 
               'opt_index_elimination',
               'opt_range_propagation'):
    g[name] = n > 0
    
  for name in ('opt_licm', 
               'opt_redundant_load_elimination', 
               'opt_stack_allocation', 
               'opt_shape_elim', 
               'stride_specialization'):
    g[name] = n > 1 
    
  for name in ('opt_loop_unrolling',):
    g[name] = n > 2 
    
opt_inline = False
opt_fusion = False
opt_index_elimination = False
opt_range_propagation = False

opt_licm = False
opt_redundant_load_elimination = False
opt_stack_allocation = False
opt_shape_elim = False

# recompile functions for distinct patterns of unit strides
stride_specialization = False

# may dramatically increase compile time
opt_loop_unrolling = False

# suspiciously complex optimizations may introduce bugs 
# TODO: comb through carefully 
opt_scalar_replacement = False
opt_copy_elimination = False
    
# run verifier after each transformation 
opt_verify = True


set_opt_level(default_opt_level)

#####################################
#            DEBUG OUTPUT           #
#####################################

# show untyped IR after it's translated from Python?
print_untyped_function = False

# show the higher level typed function after specialization?
print_specialized_function = False

# print function after all adverbs have been turned to loops
print_loopy_function = False

# show lower level typed function before
# it gets translated to LLVM?
print_lowered_function = False

# before starting function specialization, print the fn name and input types 
print_before_specialization = False

# show the input function to each transformation?
print_functions_before_transforms = [] #['Flatten', 'LowerSlices', 'LowerAdverbs', 'IndexifyAdverbs']   

# show the function produced by each transformation?
print_functions_after_transforms =  [] #['Flatten', 'LowerSlices', 'LowerAdverbs', 'IndexifyAdverbs']  

# show aliases and escape sets
print_escape_analysis = False

# how long did each transform take?
print_transform_timings = False

# print each transform's name when it runs
print_transform_names = False

# at exit, print the names of all specialized functions
print_specialized_function_names = False

#####################################
#         DESPERATE MEASURES        #
#####################################

testing_find_broken_transform = False 





from ..config import set_opt_level
from closure_specializations import print_specializations
# from decorators import jit, macro, staged_macro 

import type_conv_decls as _decls 
from decorators import jit, macro, staged_macro
from ast_conversion import translate_function_value, translate_function_ast

from run_function import run_untyped_fn, run_typed_fn, run_python_fn, specialize
from diagnose import find_broken_transform 

def typed_repr(python_fn, args, optimize = True):
  typed_fn, _ = specialize(python_fn, args)
  if optimize:
    from ..transforms import pipeline
    return pipeline.high_level_optimizations.apply(typed_fn)
  else:
    return typed_fn 



import parakeet 
from  parakeet import testing_helpers
import numpy as np 


@parakeet.jit
def setidx(x, idx, y):
    x[idx] = y

def test_set_1d_simple_slice():
    x = np.array([1,2,3,4,5,6]) 
    x2 = x.copy()
    idx = slice(2,4)
    y = [10, 20]
    print x[idx]
    x[idx] = y
    setidx(x2, idx, y)
    assert testing_helpers.eq(x, x2), "Expected %s but got %s" % (x, x2)
    
def test_set_1d_simple_slice_to_const():
    x = np.array([1,2,3,4,5,6]) 
    x2 = x.copy()
    idx = slice(2,4)
    y = 0
    x[idx] = y
    setidx(x2, idx, y)
    assert testing_helpers.eq(x, x2)

def test_set_1d_step_slice_to_const():
    x = np.array([1,2,3,4,5,6]) 
    x2 = x.copy()
    idx = slice(2,4,2)
    y = 0
    x[idx] = y
    setidx(x2, idx, y)
    assert testing_helpers.eq(x, x2)

def test_set_1d_negative_slice():
    x = np.array([1,2,3,4,5,6]) 
    x2 = x.copy()
    idx = slice(4,2,-1)
    y = [10, 20]
    x[idx] = y
    setidx(x2, idx, y)
    assert testing_helpers.eq(x, x2), "Expected %s but got %s" % (x, x2)

if __name__ == '__main__':
    testing_helpers.run_local_tests()




from .. import names 
from treelike import NestedBlocks
from ..builder import build_fn, Builder
from ..ndtypes import (ScalarT, NoneT, NoneType, ArrayT, SliceT, TupleT, make_tuple_type, 
                       Int64, PtrT, ptr_type, ClosureT, make_closure_type, FnT, StructT, 
                       TypeValueT)
from ..syntax import (Var, Attribute, Tuple, TupleProj, Closure, ClosureElt, Const,
                      Struct, Index, TypedFn, Return, Stmt, Assign, Alloc, AllocArray, 
                      ParFor, PrimCall, If, While, ForLoop, Call, Expr, 
                      IndexMap, IndexReduce, ExprStmt) 
from ..syntax.helpers import none, const_int, slice_none 

from transform import Transform

def concat(seqs):
  result = []
  for seq in seqs:
    for elt in seq:
      result.append(elt)
  return tuple(result)

def concat_args(*seqs):
  return concat(seqs)

def concat_map(f, seq):
  return concat(f(elt) for elt in seq)

def flatten_type(t):
  if isinstance(t, (ScalarT, PtrT)):
    return (t,)
  elif isinstance(t, TupleT):
    return concat(flatten_type(elt_t) for elt_t in t.elt_types)
  elif isinstance(t, (NoneT, FnT, TypeValueT)):
    return ()
  elif isinstance(t, ClosureT):
    return concat(flatten_type(elt_t) for elt_t in t.arg_types)
  elif isinstance(t, ArrayT):
    return concat_args(
      flatten_type(t.ptr_t), 
      flatten_type(t.shape_t),
      flatten_type(t.strides_t),          
      (Int64, Int64), # offset and size
      
    )
  elif isinstance(t, SliceT):
    return flatten_types( (t.start_type, t.stop_type, t.step_type) )
  else:
    assert False, "Unsupported type %s" % (t,)

def flatten_types(ts):
  return concat([flatten_type(t) for t in ts])
  
def field_pos_range(t, field, _cache = {}):
  key = (t, field)
  if key in _cache:
    return _cache[key]
  assert isinstance(t, StructT), "Expected struct got %s.%s" % (t, field)
  offset = 0
  for i, (field_name, field_t) in enumerate(t._fields_):
    n = len(flatten_type(field_t))
    if field_name == field or (isinstance(field, (int, long)) and i == field):
      result = (offset, offset+n)
      _cache[key] = result
      return result
    offset += n 
  assert False, "Field %s not found on type %s" % (field, t)

def get_field_elts(t, values, field):
  start, stop = field_pos_range(t, field)
  assert stop <= len(values), \
    "Insufficient number of flattened fields %s for %s.%s" % (values, t, field)
  return values[start:stop]

def single_type(ts):
  """
  Turn a sequence of types into a single type object
  """
  if len(ts) == 0:
    return NoneType
  elif len(ts) == 1:
    return ts[0]
  else:
    return make_tuple_type(ts)
  
def single_value(values):
  if len(values) == 0:
    return none 
  elif len(values) == 1:
    return values[0]
  else:
    t = make_tuple_type(tuple(v.type for v in values))
    return Tuple(values, type = t)
  
def mk_vars(names, types):
  """
  Combine a list of names and a list of types into a single list of vars
  """
  return [Var(name = name, type = t) for name, t in zip(names, types)]


class BuildFlatFn(Builder):
  def __init__(self, old_fn):
    Builder.__init__(self)
    self.old_fn = old_fn 
    base_name = names.original(old_fn.name)
    flat_fn_name = names.fresh("flat_" + base_name)

    self.type_env = {}
    for (name, t) in old_fn.type_env.iteritems():

      old_var = Var(name = name, type = t)
      for new_var in self.flatten_lhs_var(old_var):
  
        self.type_env[new_var.name] = new_var.type
    old_input_vars = mk_vars(old_fn.arg_names, old_fn.input_types)
        
    # a var of struct type from the old fn 
    # maps to multiple variables in the new
    # function 
    self.var_expansions = {}
    
    new_input_vars = []
    for var in old_input_vars:
      flat_vars = self.flatten_lhs_var(var)
      self.var_expansions[var.name]  = flat_vars 
      new_input_vars.extend(flat_vars)
    
    new_input_names = tuple(var.name for var in new_input_vars)
    new_input_types = tuple(var.type for var in new_input_vars)
    new_return_type =  single_type(flatten_type(old_fn.return_type))
    
    self.flat_fn = \
      TypedFn(name = flat_fn_name, 
        arg_names = new_input_names, 
        body = [], 
        type_env = self.type_env,
        input_types = new_input_types, 
        return_type = new_return_type, 
        created_by = self) 
      
    
  def run(self):
    self.flat_fn.body = self.flatten_block(self.old_fn.body)
    return self.flat_fn
  
  
  #######################
  #
  #     Helpers
  #
  #######################
  
  def flatten_block(self, stmts):
    self.blocks.push()
    for stmt in stmts:
      result = self.flatten_stmt(stmt)
      if result is None:
        continue
      if not isinstance(result, (list,tuple)):
        result = [result]
      for new_stmt in result:
        assert isinstance(new_stmt, Stmt), "Expected statement, got %s" % (new_stmt,)
        self.blocks.append(new_stmt)
    return self.blocks.pop()
  
  
  
  #######################
  #
  #     Statements
  #
  #######################
  
  
  #def bind_name(self):
  #def bind_var(self, lhs, rhs):
  
  def flatten_Assign(self, stmt):
    c = stmt.lhs.__class__
    rhs = self.flatten_expr(stmt.rhs)
    if c is Var:
      
      lhs_vars = self.flatten_lhs_var(stmt.lhs)
      self.var_expansions[stmt.lhs.name] = lhs_vars 
      if isinstance(rhs, Expr):
        if len(lhs_vars) == 1:
          return [Assign(lhs_vars[0], rhs)]
        else:
          # the IR doesn't allow for multiple assignment 
          # so we fake it with tuple literals
          return [Assign(self.tuple(lhs_vars), rhs)]
      assert isinstance(rhs, (list,tuple))
      assert len(lhs_vars) == len(rhs), \
        "Mismatch between LHS %s and RHS %s : %s => %s in stmt %s" % \
        (lhs_vars, stmt.rhs, stmt.rhs.type, rhs, stmt)
      result = []
      for var, value in zip(lhs_vars, rhs):
        result.append(Assign(var, value))
      return result 

    elif c is Index:
      array_t = stmt.lhs.value.type 
      if isinstance(array_t, PtrT):
        return stmt  
      indices = self.flatten_expr(stmt.lhs.index)
      values = self.flatten_expr(stmt.lhs.value)
      data = get_field_elts(array_t, values, 'data')[0]
      shape = get_field_elts(array_t, values, 'shape')
      strides = get_field_elts(array_t, values, 'strides')
      offset = get_field_elts(array_t, values, 'offset')[0]
      n_dims = len(strides)
      n_indices = len(indices)
      assert n_dims == n_indices, \
        "Expected %d indices but only got %d in %s" % (n_dims, n_indices, stmt)
      for idx, stride in zip(indices, strides):
        offset = self.add(offset, self.mul(idx, stride))

      stmt.lhs = self.index(data, offset, temp=False)
      stmt.rhs = rhs[0]
      return [stmt]
    else:
      assert False, "LHS not supported in flattening: %s" % stmt 
  
  def enter_branch(self, phi_nodes):
    for (k, (left, _)) in phi_nodes.iteritems():
      self.var_expansions[k] = self.flatten_lhs_name(k, left.type)
  
  def flatten_merge(self, phi_nodes):
    result = {}
    for (k, (left, right)) in phi_nodes.iteritems():
      t = left.type
      assert right.type == t 
      if isinstance(t, (ScalarT, PtrT)):
        result[k] = (self.flatten_scalar_expr(left), self.flatten_scalar_expr(right))
      elif isinstance(t, (FnT, NoneT)):
        continue 
      else:
      
        fields = self.var_expansions[k]
        flat_left = self.flatten_expr(left)
        flat_right = self.flatten_expr(right)
        assert len(fields) == len(flat_left)
        assert len(fields) == len(flat_right)
        for i, var in enumerate(fields):
          result[var.name] = (flat_left[i], flat_right[i])
    return result 
   
  def flatten_ForLoop(self, stmt):
    self.enter_branch(stmt.merge)
    var = self.flatten_scalar_lhs_var(stmt.var)
    start = self.flatten_scalar_expr(stmt.start)
    stop = self.flatten_scalar_expr(stmt.stop)
    step = self.flatten_scalar_expr(stmt.step)
    body = self.flatten_block(stmt.body)
    merge = self.flatten_merge(stmt.merge)
    return ForLoop(var, start, stop, step, body, merge)
  
  def flatten_While(self, stmt):
    self.enter_branch(stmt.merge)
    cond = self.flatten_scalar_expr(stmt.cond)
    body = self.flatten_block(stmt.body)
    merge = self.flatten_merge(stmt.merge)
    return While(cond, body, merge)
     
  
  def flatten_If(self, stmt):
    self.enter_branch(stmt.merge)
    cond = self.flatten_scalar_expr(stmt.cond)
    true = self.flatten_block(stmt.true)
    false = self.flatten_block(stmt.false)
    merge = self.flatten_merge(stmt.merge)
    assert merge is not None
    return If(cond, true, false, merge = merge)
  
  def flatten_ExprStmt(self, stmt):
    return ExprStmt(value = self.flatten_expr(stmt.value))
   
  def flatten_ParFor(self, stmt):
    new_fn, closure_elts = self.flatten_fn(stmt.fn)
    closure = self.closure(new_fn, closure_elts)
    bounds = single_value(self.flatten_expr(stmt.bounds))
    return ParFor(fn = closure, bounds = bounds)
    
  def flatten_Return(self, stmt):
    return Return(single_value(self.flatten_expr(stmt.value)))
  
  def flatten_Comment(self, stmt):
    return stmt 
  
  def flatten_stmt(self, stmt):
    method_name = "flatten_%s" % stmt.__class__.__name__
    return getattr(self, method_name)(stmt)
  
  #######################
  #
  #     Expressions 
  #
  #######################
  
  def flatten_expr(self, expr):
    method_name = "flatten_%s" % expr.__class__.__name__
    return getattr(self, method_name)(expr)
  
  def flatten_expr_list(self, exprs):
    return concat_map(self.flatten_expr, exprs)
  
  def flatten_scalar_expr(self, expr):
    """
    Give me back a single expression instead of a list 
    """
    flat_exprs = self.flatten_expr(expr)
    assert len(flat_exprs) == 1
    return flat_exprs[0]
    
  def flatten_scalar_expr_list(self, exprs):
    assert isinstance(exprs, (list, tuple)), "Expected list, got %s" % (exprs,)
    return [self.flatten_scalar_expr(e) for e in  exprs]
  
  def flatten_Const(self, expr):
    if isinstance(expr.type, NoneT):
      return ()
    return (expr,)
  
  def flatten_fn(self, closure):
    fn = self.get_fn(closure)
    import pipeline
    fn = pipeline.indexify.apply(fn)
    flat_fn = build_flat_fn(fn)
    flat_closure_args = self.flatten_expr(closure)
    return flat_fn, flat_closure_args
  
  
  def flatten_Call(self, expr):
    flat_fn, flat_closure_args = self.flatten_fn(expr.fn)
    flat_args = self.flatten_expr_list(expr.args)
    args = tuple(flat_closure_args) + tuple(flat_args)
    return Call(flat_fn, args, type = flat_fn.return_type)
  
  def flatten_Cast(self, expr):
    return [expr]
  
  def flatten_UntypedFn(self, expr):
    return []
  
  def flatten_TypedFn(self, expr):
    return []
    
  def flatten_Var(self, expr):
    if isinstance(expr.type, (ScalarT, PtrT)):
      return (expr,)
    elif isinstance(expr.type, (FnT, NoneT)):
      return ()
    else:
      name = expr.name
      assert name in self.var_expansions, "No fields known for %s : %s" % (expr, expr.type) 
      return self.var_expansions[name]
    
  def flatten_Tuple(self, expr):
    return self.flatten_expr_list(expr.elts)
  
  def flatten_field(self, struct, field):
    elts = self.flatten_expr(struct)
    start, stop = field_pos_range(struct.type, field)
    return elts[start:stop]
   
  def flatten_TupleProj(self, expr):
    result = self.flatten_field(expr.tuple, expr.index)
    return result
    
  def flatten_Closure(self, expr):
    return self.flatten_expr_list(expr.args)
    
  def flatten_ClosureElt(self, expr):
    return self.flatten_field(expr.closure, expr.index)
  
  def flatten_Attribute(self, expr):
    return self.flatten_field(expr.value, expr.name)
  
  def flatten_Select(self, expr):
    cond = self.flatten_scalar_expr(expr.cond)
    true_values = self.flatten_expr(expr.true_value)
    false_values = self.flatten_expr(expr.false_value)
    assert len(true_values) == len(false_values)
    return [self.select(cond,t,f) for t,f in zip(true_values, false_values)]
    
  def flatten_Alloc(self, expr):
    count_exprs = self.flatten_expr(expr.count)
    assert len(count_exprs) == 1
    return Alloc(count = count_exprs[0], elt_type = expr.elt_type, type = expr.type)
    
  def flatten_Array(self, expr):
    assert False, "Array node should be an explicit allocation by now"
    # or, if we flatten structured elts, maybe we should handle it here?
  
    
  def flatten_Index(self, expr):
    t = expr.value.type 
    if isinstance(t, PtrT):
      return [expr]   
    assert isinstance(t, ArrayT), "Expected Index to take array, got %s" % (expr.type,)
    array_fields = self.flatten_expr(expr.value)
    data_fields = get_field_elts(t, array_fields, 'data')
    shape = get_field_elts(t, array_fields, 'shape')
    strides = get_field_elts(t, array_fields, 'strides')
    offset = get_field_elts(t, array_fields, 'offset')[0]
    
    indices = self.flatten_expr(expr.index)
    
    n_indices = len(indices)
    n_strides = len(strides)
    assert n_indices == n_strides, \
      "Not supported:  indices vs. dimensions: %d != %d in %s" % (n_indices, n_strides, expr)


    # fast-path for the common case when we're indexing
    # by all scalars to retrieve a scalar result
    #if syntax.helpers.all_scalars(indices):
    for i, idx in enumerate(indices):
      offset = self.add(offset, self.mul(idx, strides[i]))
    return [self.index(data_fields[0], offset)]
     
  
  def flatten_PrimCall(self, expr):
    args = self.flatten_scalar_expr_list(expr.args)
    return [PrimCall(prim = expr.prim, args = args, type = expr.type)]
  
  def flatten_Slice(self, expr):
    return self.flatten_expr_list([expr.start, expr.stop, expr.step])
  
  def flatten_Len(self, expr):
    assert False, "Not implemented" 
  
  def flatten_ConstArray(self, expr):
    assert False, "Not implemented" 
  
  def flatten_ConstArrayLike(self, expr):
    assert False, "Not implemented" 
  
  def flatten_Range(self, expr):
    assert False, "Not implemented" 
  
  def strides_from_shape_elts(self, shape_elts):
    strides = [const_int(1)]
    for dim in reversed(shape_elts[1:]):
      strides = [self.mul(strides[0], dim)] + strides
    return strides
  
  def flatten_AllocArray(self, expr):
    nelts = const_int(1)
    shape_elts = self.flatten_expr(expr.shape)
    for dim in shape_elts:
      nelts = self.mul(nelts, dim)
    ptr = Alloc(elt_type = expr.elt_type, count = nelts, type = ptr_type(expr.elt_type))
    stride_elts = self.strides_from_shape_elts(shape_elts)
    return (ptr,) + tuple(shape_elts) + tuple(stride_elts) + (self.int(0), nelts)
  
  def flatten_ArrayView(self, expr):
    data = self.flatten_expr(expr.data)
    shape = self.flatten_expr(expr.shape)
    strides = self.flatten_expr(expr.strides)
    offset = self.flatten_expr(expr.offset)
    size = self.flatten_expr(expr.size)
    return data + shape + strides + offset + size
    
  
  def flatten_Ravel(self, expr):
    assert False, "Not implemented" 
  
  def flatten_Reshape(self, expr):
    assert False, "Not implemented" 
  
  def flatten_Shape(self, expr):
    assert False, "Not implemented" 
  
  def flatten_Strides(self, expr):
    assert False, "Not implemented" 
  
  def flatten_Transpose(self, expr):
    assert False, "Not implemented" 
  
  def flatten_Where(self, expr):
    assert False, "Not implemented" 
     
   
    
  #######################
  #
  # Adverbs
  #
  #######################
  def flatten_Map(self, expr):
    assert False, "Unexpected Map encountered during flattening, should be IndexScan"
  
  def flatten_Reduce(self, expr):
    assert False, "Unexpected Reduce encountered during flattening, should be IndexScan"
  
  def flatten_Scan(self, expr):
    assert False, "Unexpected Scan encountered during flattening, should be IndexScan"
  
  def flatten_IndexMap(self, expr):
    #fn, closure_args = self.flatten_fn(expr.fn)
    #shape_elts = self.flatten_expr(expr.shape)
    # import pdb; pdb.set_trace()
    #return IndexMap(fn = self.closure(fn, closure_args), shape = self.tuple(shape_elts), type = None)
    assert False, "Unexpected IndexMap, should have been turned into ParFor before flattening"
    
  
  def flatten_OuterMap(self, expr):
    assert False, "Unexpected OuterMap, should have been turned into ParFor before flattening"
  
  def flatten_IndexReduce(self, expr):
    # assert isinstance(expr.type, ScalarT), "Non-scalar reductions not yet implemented"
    fn, fn_args = self.flatten_fn(expr.fn)
    fn = self.closure(fn, fn_args)
    combine, combine_args = self.flatten_fn(expr.combine)
    combine = self.closure(combine, combine_args)
    shape = self.tuple(self.flatten_expr(expr.shape))
    
    init = self.flatten_expr(expr.init)
    t = flatten_type(expr.type)
    if len(t) == 1:
      init = init[0]
      t = t[0]
    else:
      init = self.tuple(init)
      t = make_tuple_type(t)
    
    result =  IndexReduce(fn = fn, combine = combine, shape = shape, type = t, init = init)
    return [result]
     
      
  def flatten_IndexScan(self, expr):
    assert False, "IndexScan Not implemented" 
  
  def flatten_lhs_name(self, name, t):
    if isinstance(t, (ScalarT, PtrT)):
      return [Var(name = name, type = t)]
    elif isinstance(t, (NoneT, FnT, TypeValueT)):
      return []
    elif isinstance(t, SliceT):
      start = Var(name = "%s_start" % name, type = t.start_type)
      stop = Var(name = "%s_stop" % name, type = t.stop_type)
      step = Var(name = "%s_step" % name, type = t.step_type)
      field_vars = [start, stop, step]
    elif isinstance(t, ClosureT):
      field_vars = [Var(name = "%s_closure_elt%d" % (name,i) , type = t) 
                    for i,t in enumerate(t.arg_types)]
    elif isinstance(t, TupleT):
      field_vars = [Var(name = "%s_elt%d" % (name, i), type = t) 
                    for i,t in enumerate(t.elt_types)]
    elif isinstance(t, ArrayT):
      data = Var(name = "%s_data" % name, type = t.ptr_t)
      shape = Var(name = "%s_shape" % name, type = t.shape_t)
      strides = Var(name = "%s_strides" % name, type = t.strides_t)
      offset = Var(name = "%s_offset" % name, type = Int64)
      nelts = Var(name = "%s_nelts" % name, type = Int64)
      field_vars = [data, shape, strides, offset, nelts]
    else:
      assert False, "Unsupport type %s" % (t,)
    return self.flatten_lhs_vars(field_vars)
  
  def flatten_lhs_var(self, old_var):
    t = old_var.type 
    if isinstance(t, (PtrT, ScalarT)):
      return [old_var]
    name = old_var.name 
    return self.flatten_lhs_name(name, t)
  
  def flatten_lhs_vars(self, old_vars):
    return concat_map(self.flatten_lhs_var, old_vars)
  
  def flatten_scalar_lhs_var(self, old_var):
    lhs_vars = self.flatten_lhs_var(old_var)
    assert len(lhs_vars) == 1
    return lhs_vars[0]
  
def build_flat_fn(old_fn, _cache = {}):
  key = old_fn.cache_key
  if key in _cache:
    return _cache[key]
  flat_fn = BuildFlatFn(old_fn).run()
  _cache[key] = flat_fn
  _cache[flat_fn.cache_key] = flat_fn
  return flat_fn
  
class Flatten(Transform):
  
  def unbox_var(self, var):
    t = var.type 
    if isinstance(t, (FnT, NoneT, TypeValueT)):
      return []
    elif isinstance(t, (PtrT, ScalarT)):
      return [var]
    elif isinstance(t, ArrayT):
      # for structured arrays, should this be a tuple? 
      data = self.attr(var, 'data', name = var.name + "_data")
      shape = self.attr(var, 'shape', name = var.name + "_shape")
      strides = self.attr(var, 'strides', name = var.name + "_strides")
      offset = self.attr(var, 'offset', name = var.name + "_offset")
      size = self.attr(var, 'size', name = var.name + "_size")
      return self.unbox_vars([data, shape, strides, offset, size])
    elif isinstance(t, SliceT):
      start = self.attr(var, 'start')
      stop = self.attr(var, 'stop')
      step = self.attr(var, 'step')
      return self.unbox_vars([start, stop, step])
    elif isinstance(t, ClosureT):
      closure_elts = [self.closure_elt(var, i, name = var.name + "_closure_elt%d" % i)
                      for i in xrange(len(t.arg_types))]
      return self.unbox_vars(closure_elts)
    elif isinstance(t, TupleT):
      tuple_elts = [self.assign_name(self.tuple_proj(var, i), name = var.name + "_elt%d" % i)
                      for i in xrange(len(t.elt_types))]
      return self.unbox_vars(tuple_elts)
    else:
      assert False, "Unsupported type %s" % (t,)
  
  def unbox_vars(self, exprs):
    return concat_map(self.unbox_var, exprs)
  
  def to_seq(self, expr):
    if isinstance(expr.type, TupleT):
      return self.tuple_elts(expr)
    elif isinstance(expr.type, (FnT, NoneT)):
      return []
    else:
      return [expr]
  
  def box(self, t, elts):
    if isinstance(t, NoneT):
      assert len(elts) == 0, "Expected 0 values for None, got %s" % (elts,)
      return none 
    elif isinstance(t, ScalarT):
      assert len(elts) == 1
      return elts[0]
    elif isinstance(t, SliceT):
      assert len(elts) == 3
      start, stop, step = elts
      return self.slice_value(start, stop, step)
    elif isinstance(t, ArrayT):
      data = get_field_elts(t, elts, 'data')[0]
      shape = self.tuple(get_field_elts(t, elts, 'shape'))
      strides = self.tuple(get_field_elts(t, elts, 'strides'))
      offset = get_field_elts(t, elts, 'offset')[0]
      nelts = get_field_elts(t, elts, 'size')[0]
      return self.array_view(data, shape, strides, offset, nelts)
    elif isinstance(t, TupleT):
      boxed_elts = []
      for i, elt_t in enumerate(t.elt_types):
        elt = self.box(elt_t, get_field_elts(t, elts, i))
        boxed_elts.append(elt)
      return self.tuple(boxed_elts)
    elif isinstance(t, ClosureT):
      assert False, "Not implemented: ClosureT" 
    elif isinstance(t, FnT):
      assert False, "Not implemented: FnT" 
  
  def transform_block(self, stmts):
    return stmts
    
  def pre_apply(self, old_fn, _cache = {}):
    key = old_fn.cache_key
    if key  in _cache:
      return _cache[key]
    
    flat_fn = build_flat_fn(old_fn)
    
    flat_fn.created_by = old_fn.created_by
    flat_fn.transform_history = old_fn.transform_history.copy()
    
    input_vars = mk_vars(old_fn.arg_names, old_fn.input_types)
    self.blocks.push()
    unboxed_inputs = self.unbox_vars(input_vars)
    assert len(unboxed_inputs) == len(flat_fn.input_types)
    unboxed_result = self.call(flat_fn, unboxed_inputs, name = "unboxed_result")
    unboxed_elts = self.to_seq(unboxed_result)
    boxed_result = self.box(old_fn.return_type, unboxed_elts)
    self.return_(boxed_result)
    old_fn.body = self.blocks.pop()
    _cache[key] = old_fn
    _cache[flat_fn.cache_key] = flat_fn 
    return old_fn
  


from ..c_backend import PyModuleCompiler, FlatFnCompiler

class CudaFlatFnCompiler(FlatFnCompiler):
  def visit_ParFor(self, stmt):
    pass 
   
  def visit_IndexMap(self, expr):
    pass
  
  def visit_IndexReduce(self, expr):
    pass 
  
  def visit_IndexScan(self, expr):
    pass  
  

class CudaModuleCompiler(PyModuleCompiler, CudaFlatFnCompiler):
  pass  

def compile_entry(fn):
  pass 


from ..c_backend import PyModuleCompiler, FlatFnCompiler

class MulticoreFlatFnCompiler(FlatFnCompiler):
  def visit_ParFor(self, stmt):
    bounds = self.tuple_to_var_list(stmt.bounds)
    loop_var_names = ["i", "j", "k", "l", "ii", "jj", "kk", "ll"]
    n_vars = len(bounds)
    assert n_vars <= len(loop_var_names)
    loop_vars = [self.fresh_var("int64_t", loop_var_names[i]) for i in xrange(n_vars)]
    
    omp = "#pragma omp parallel for" 
    fn_name = self.get_fn(stmt.fn)
    closure_args = self.get_closure_args(stmt.fn)
    combined_args = tuple(closure_args) + tuple(loop_vars)
    arg_str = ", ".join(combined_args)
    body = "%s(%s);" % (fn_name, arg_str)    
    return omp + self.build_loops(loop_vars, bounds, body) 
  
  def visit_IndexMap(self, expr):
    pass
  
  def visit_IndexReduce(self, expr):
    pass 
  
  def visit_IndexScan(self, expr):
    pass 

class MulticoreModuleCompiler(PyModuleCompiler, MulticoreFlatFnCompiler):
  pass  

def compile_entry(fn):
  pass 


import numpy as np


from parakeet.testing_helpers import run_local_tests, expect

mat_shape = (2,3)
vec_len = np.prod(mat_shape)

int_vec = np.arange(vec_len)
float_vec = np.sqrt(int_vec)
bool_vec = int_vec % 2 

vectors = [int_vec, float_vec, bool_vec]
matrices = [np.reshape(vec, mat_shape) for vec in vectors]

def test_dot_vv():
  for x in vectors:
    for y in vectors:
      res = np.dot(x,y)
      expect(np.dot, [x,y], res)

def test_dot_mm():
  for X in matrices:
    for Y in matrices:
      res = np.dot(X, Y.T)
      expect(np.dot, [X,Y.T], res)

if __name__ == '__main__':
    run_local_tests()



import numpy as np 
from parakeet import jit, testing_helpers 

@jit 
def true_divided(x):
    return True / x

def test_true_divided_python_scalars():
    testing_helpers.expect(true_divided, [True], [True])
    testing_helpers.expect(true_divided, [1], [1])
    testing_helpers.expect(true_divided, [1.0], [1.0])
    testing_helpers.expect(true_divided, [2], [0])
    testing_helpers.expect(true_divided, [2.0], [0.5])


def test_true_divided_uint8():
    testing_helpers.expect(true_divided, [np.uint8(1)], 1)
    testing_helpers.expect(true_divided, [np.uint8(2)], 0)

if __name__ == '__main__':
    testing_helpers.run_local_tests()



import collections 
import distutils 
import imp 
import numpy.distutils as npdist 
import numpy.distutils.system_info as np_sysinfo 
import os
import platform
import subprocess  
import time 

from tempfile import NamedTemporaryFile

import config 



CompiledPyFn = collections.namedtuple("CompiledPyFn",
                                      ("c_fn", 
                                       "src_filename", 
                                       "module", 
                                       "shared_filename", 
                                       "object_filename",
                                       "src", 
                                       "fn_name",
                                       "fn_signature"))

CompiledObject = collections.namedtuple("CompiledObject", 
                                        ("src",
                                         "src_filename",  
                                         "object_filename", 
                                         "fn_name",
                                         "fn_signature"))

  
c_headers = ["stdint.h",  "math.h",  "signal.h"]
core_python_headers = ["Python.h"]
numpy_headers = ['numpy/arrayobject.h', 'numpy/arrayscalars.h']

python_headers = core_python_headers + numpy_headers 

cpp_defs = [] #"#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION"]

config_vars = distutils.sysconfig.get_config_vars()


def get_compiler(_cache = {}):
  if config.pure_c in _cache:
    return _cache[config.pure_c]
  for compiler in [ ('gcc' if config.pure_c else 'g++'), 'icc', 'clang']:
    path = distutils.spawn.find_executable(compiler)
    if path:
      _cache[config.pure_c] = path
      return path 
  assert False, "No compiler found!"
    

def get_source_extension():
  return ".c" if config.pure_c else ".cpp"

object_extension = ".o"
shared_extension = np_sysinfo.get_shared_lib_extension(True)


mac_os = platform.system() == 'Darwin'
windows = platform.system() == 'Windows'
if mac_os:
  python_lib_extension = '.dylib'
else:
  python_lib_extension = '.so'

python_include_dirs = [distutils.sysconfig.get_python_inc()]

numpy_include_dirs = npdist.misc_util.get_numpy_include_dirs()
include_dirs = python_include_dirs + numpy_include_dirs 


def use_openmp(compiler):
  return not config.debug and config.use_openmp and compiler in ('gcc', 'g++', 'icc')

def get_opt_flags():
  opt_flags = ['-O3', '-msse2']
  if config.fast_math:
    opt_flags.append('-ffast-math')
  return opt_flags 

def get_compiler_flags(compiler):
  compiler_flags = ['-I%s' % path for path in include_dirs]
  compiler_flags.extend(['-fPIC', '-Wall', '-Wno-unused-variable'])
  if config.debug:
    compiler_flags.extend(['-g', '-O0'])
  else:
    compiler_flags.extend(get_opt_flags())

  if not config.pure_c: 
    compiler_flags.extend(['-fpermissive'])
  
  if use_openmp(compiler):
    compiler_flags.append('-fopenmp')  
  return compiler_flags   

python_lib_dir = distutils.sysconfig.get_python_lib() + "/../../"
python_version = distutils.sysconfig.get_python_version()
#python_lib = "python%s" % python_version
#python_lib_full = 'lib%s%s' % (python_lib, python_lib_extension)

def get_linker_flags(compiler):
  linker_flags = ['-shared'] + ['-lm']
  
  if mac_os:
    linker_flags.append("-headerpad_max_install_names")
    linker_flags.append("-undefined dynamic_lookup")

  if use_openmp(compiler):
    linker_flags.append('-fopenmp')             
  return linker_flags 




def create_source_file(src, 
                         fn_name = None, 
                         src_filename = None, 
                         forward_declarations = [],
                         extra_function_sources = [], 
                         extra_headers = [], 
                         print_source = None):
  
  if print_source is None: 
    print_source = config.print_module_source

  if fn_name is None:
    prefix = "parakeet_"
  else:
    prefix = "parakeet_%s_" % fn_name
  if src_filename is None:
    src_file = NamedTemporaryFile(suffix = get_source_extension(),  
                                  prefix =  prefix, 
                                  delete = False, 
                                  mode = 'w')
    src_filename = src_file.name 
  else:
    src_file = open(src_filename, 'w')
  
  for d in cpp_defs:
    src_file.write(d)
    src_file.write("\n")
  
  for header in extra_headers + c_headers:
    src_file.write("#include <%s>\n" % header)
  
  for decl in set(forward_declarations):
    decl = decl.strip()
    if not decl.endswith(";"):
      decl += ";"
    decl += "\n"
    src_file.write(decl)
  
  for other_fn_src in extra_function_sources:
    src_file.write(other_fn_src)
    src_file.write("\n")
      
  src_file.write(src)
  src_file.close()
  if print_source: print subprocess.check_output(['cat', src_filename])
  return src_file 

def run_cmd(cmd, env = None, label = ""):
  if config.print_commands: print " ".join(cmd)
  if config.print_command_elapsed_time: t = time.time()
  with open(os.devnull, "w") as fnull:
    with NamedTemporaryFile(prefix="parakeet_compile_err", mode = 'r+') as err_file:
      try:  
        subprocess.check_call(cmd, stdout = fnull, stderr = err_file, env = env)
      except:
        print "Parakeet encountered error(s) during compilation: "
        print err_file.read()
        raise 
    
  if config.print_command_elapsed_time: 
    if label:
      print "%s, elapsed time: %0.4f" % (label, time.time() - t)
    else:
      print "Elapsed time:", time.time() - t 

def compile_object(src, 
                   fn_name = None,  
                   fn_signature = None,
                   src_filename = None, 
                   forward_declarations = [],
                   extra_function_sources = [], 
                   extra_headers = python_headers, 
                   extra_objects = [], 
                   print_source = None, 
                   print_commands = None):
  if print_source is None: print_source = config.print_module_source
  if print_commands is None: print_commands = config.print_commands
  
  src_file = create_source_file(src, 
                                fn_name = fn_name, 
                                src_filename = src_filename, 
                                forward_declarations = forward_declarations,
                                extra_function_sources = extra_function_sources,  
                                extra_headers = extra_headers,
                                print_source = print_source)
  src_filename = src_file.name
  object_name = src_filename.replace(get_source_extension(), object_extension)
  compiler = get_compiler()
  compiler_flags = get_compiler_flags(compiler)
  compiler_cmd = [compiler] + compiler_flags + ['-c', src_filename, '-o', object_name]
  run_cmd(compiler_cmd, label = "Compile source")
  return CompiledObject(src = src, 
                        src_filename = src_filename, 
                        object_filename = object_name, 
                        fn_name = fn_name, 
                        fn_signature = fn_signature)
  
  
def compile_module(src, 
                     fn_name,
                     fn_signature = None,  
                     src_filename = None,
                     forward_declarations = [],
                     extra_function_sources = [], 
                     extra_headers = [],  
                     extra_objects = [],
                     print_source = None, 
                     print_commands = None):
  
  if print_source is None:
    print_source = config.print_module_source
  if print_commands is None:
    print_commands = config.print_commands

  src += """
    static PyMethodDef %(fn_name)sMethods[] = {
      {"%(fn_name)s",  %(fn_name)s, METH_VARARGS,
       "%(fn_name)s"},

      {NULL, NULL, 0, NULL}        /* Sentinel */
    };
  
    PyMODINIT_FUNC
    init%(fn_name)s(void)
    {
      //Py_Initialize();
      Py_InitModule("%(fn_name)s", %(fn_name)sMethods);
      import_array();
    }
    """ % locals()  
  

  compiled_object = compile_object(src, 
                                   fn_name,
                                   src_filename  = src_filename, 
                                   forward_declarations = forward_declarations,
                                   extra_function_sources = extra_function_sources, 
                                   extra_headers = python_headers + extra_headers,  
                                   extra_objects = extra_objects,
                                   print_source = print_source, 
                                   print_commands = print_commands)
  
  src_filename = compiled_object.src_filename
  object_name = compiled_object.object_filename
  
  shared_name = src_filename.replace(get_source_extension(), shared_extension)
  compiler = get_compiler()
  linker_flags = get_linker_flags(compiler)
  linker_cmd = [compiler] + linker_flags + [object_name] + list(extra_objects) + ['-o', shared_name]

  env = os.environ.copy()
  env["LD_LIBRARY_PATH"] = python_lib_dir
  run_cmd(linker_cmd, env = env, label = "Linking")
  
  #if mac_os:
  #  # Annoyingly have to patch up the shared library to point to the correct Python
  #  change_cmd = ['install_name_tool', '-change', '%s' % python_lib_full, 
  #                python_lib_dir + "/%s" % python_lib_full, 
  #                shared_name]
  #  if print_commands: print " ".join(change_cmd)
  #  subprocess.check_output(change_cmd)
  # delete the .o file since we don't need it anymore 
  # os.remove(object_name)
  
  if print_commands:
    print "Loading newly compiled extension module %s..." % shared_name
  module =  imp.load_dynamic(fn_name, shared_name)
  
  #on a UNIX-style filesystem it should be OK to delete a file while it's open
  #since the inode will just float untethered from any name
  #If we ever support windows we should find some other way to delete the .dll 
  
  
  c_fn = getattr(module,fn_name)
  
  if config.delete_temp_files:
    os.remove(src_filename)
    os.remove(object_name)
    # window's can't just untether inodes like a UNIX
    # ...have to eventually think of a plan to clean these things up
    if not windows: os.remove(shared_name)
    
  compiled_fn = CompiledPyFn(c_fn = c_fn, 
                             module = module, 
                             shared_filename =  shared_name,
                             object_filename = object_name, 
                             src = src, 
                             src_filename = src_filename,
                             fn_name = fn_name, 
                             fn_signature = fn_signature)
  return compiled_fn



#!/usr/bin/env python

from setuptools import setup, Extension, find_packages
import os 
import sys

import parakeet.version

setup(
    name="parakeet",
    description="Runtime compiler for numerical Python.",
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.md')).read(),
    classifiers=['Development Status :: 3 - Alpha',
                 'Topic :: Software Development :: Libraries',
                 'License :: OSI Approved :: BSD License',
                 'Intended Audience :: Developers',
                 'Programming Language :: Python :: 2.7',
                 ],
    author="Alex Rubinsteyn",
    author_email="alexr@cs.nyu.edu",
    license="BSD",
    version=parakeet.version.__version__,
    url=parakeet.version.__website__, 
    packages=find_packages() + ['parakeet.test', 'parakeet.benchmarks', 'parakeet.examples'], 
    package_dir={ 
                  'parakeet.benchmarks' : './benchmarks', 
                  'parakeet.examples' : './examples', 
                  'parakeet.test' : './test' 
                },
    install_requires=[
      'numpy>=1.7', 
      'scipy',
      'treelike'
      # LLVM is optional as long as you use the C backend 
      # 'llvmpy'   
    ])



import numpy as np
import parakeet
from parakeet.testing_helpers import assert_eq_arrays, run_local_tests

@parakeet.jit
def each_sum(X):
  return parakeet.each(parakeet.sum, X)

m = 2000
n = 1000
X = np.random.random((m,n))

def test_each_row_sum():
  assert_eq_arrays(np.sum(X, axis = 1), each_sum(X), "sum rows")

@parakeet.jit
def col_sum(X):
  return np.sum(X, axis = 0)

def test_each_col_sum():
  assert_eq_arrays(np.sum(X, axis = 0), col_sum(X), "sum cols")

if __name__ == '__main__':
  run_local_tests()



from .. import names 

from c_types import to_ctype
from reserved_names import is_reserved


class BaseCompiler(object):
  
  def __init__(self):
    self.blocks = []
    self.name_versions = {}
    self.name_mappings = {}
  
  
  def visit_expr(self, expr):
    expr_class_name = expr.__class__.__name__
    method_name = "visit_" + expr_class_name
    assert hasattr(self, method_name), "Unsupported expression %s" % expr_class_name  
    result = getattr(self, method_name)(expr)
    assert result is not None, \
      "Compilation method for expression %s returned None, expected code string" % expr_class_name
    return result 
      
  def visit_expr_list(self, exprs):
    return [self.visit_expr(e) for e in exprs]
  
  def breakpoint(self):
    self.append("raise(SIGINT);")
    
  def visit_stmt(self, stmt):
    stmt_class_name = stmt.__class__.__name__
    method_name = "visit_" + stmt_class_name
    assert hasattr(self, method_name), "Unsupported statemet %s" % stmt_class_name  
    result = getattr(self, method_name)(stmt)
    assert result is not None, "Compilation method for statement %s return None" % stmt_class_name
    return result 
  
  def push(self):
    self.blocks.append([])
  
  def pop(self):
    stmts = self.blocks.pop()
    return "  " + self.indent("\n".join("  " + stmt for stmt in stmts))
  
  def indent(self, block_str):
    return block_str.replace("\n", "\n  ")
  
  def append(self, stmt):
    stripped = stmt.strip()
    
    assert len(stripped) == 0 or \
      ";" in stripped or \
      stripped.startswith("//") or \
      stripped.startswith("/*"), "Invalid statement: %s" % stmt
    self.blocks[-1].append(stmt)
  
  def newline(self):
    self.append("\n")
    
  def comment(self, text):
    self.append("// %s" % text)
  
  def printf(self, fmt, *args):
    
    result = 'printf("%s\\n"' % fmt
    if len(args) > 0:
      result = result + ", " + ", ".join(str(arg) for arg in args)
    self.append( result + ");" )
  
  def fresh_name(self, prefix):
    prefix = names.original(prefix)
    
    prefix = prefix.replace(".", "")
  
    version = self.name_versions.get(prefix, 1)
    self.name_versions[prefix] = version + 1
    
    # not valid chars!
    if not any(c.isalpha() for c in prefix):
      prefix = "temp" + prefix 
      
    if version == 1 and not is_reserved(prefix):
      return prefix 
    elif prefix[-1] != "_":
      return "%s_%d" % (prefix, version)
    else:
      return prefix + str(version)
    
  def fresh_var(self, t, prefix = None, init = None):
    if prefix is None:
      prefix = "temp"
    name = self.fresh_name(prefix)
    if isinstance(t, str):
      t_str = t
    else:
      t_str = to_ctype(t)
    if init is None:
      self.append("%s %s;" % (t_str, name))
    else:
      self.append("%s %s = %s;" % (t_str, name, init))
    return name
  
  def assign(self, name, rhs):
    self.append("%s = %s;" % (name, rhs))
  
  def name(self, ssa_name, overwrite = False):
    """
    Convert from ssa names, which might have large version numbers and contain 
    syntactically invalid characters to valid local C names
    """
    if ssa_name in self.name_mappings and not overwrite:
      return self.name_mappings[ssa_name]
    prefix = names.original(ssa_name)
    prefix = prefix.replace(".", "")

    name = self.fresh_name(prefix)
    self.name_mappings[ssa_name] = name 
    return name
   
  def return_if_null(self, obj):
    self.append("if (!%s) { return NULL; }" % obj)


import config

from version import *
from ndtypes import * 

from analysis import SyntaxVisitor, verify

from builder import Builder, build_fn, mk_identity_fn, mk_cast_fn

from transforms import clone_function, Transform


  
from lib import * 
from prims import *

from frontend import jit, macro, run_python_fn, run_untyped_fn, run_typed_fn
from frontend import typed_repr, specialize, find_broken_transform


